@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2014 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2020 Codership Oy <info@codership.com>
  *
  * $Id$
  */
@@ -136,20 +136,45 @@ struct gcs_conn
     /* Flow Control */
     gu_mutex_t   fc_lock;
     uint32_t     conf_id;             // configuration ID
-    long         stop_sent;           // how many STOPs - CONTs were sent
+    int          stop_sent_;          // how many STOPs - CONTs were sent
+    int          stop_sent()
+    {
+#ifdef GU_DEBUG_MUTEX
+        assert(gu_mutex_owned(&fc_lock));
+#endif
+        return stop_sent_;
+    }
+    void         stop_sent_inc(int val)
+    {
+#ifdef GU_DEBUG_MUTEX
+        assert(gu_mutex_owned(&fc_lock));
+#endif
+        stop_sent_ += val;
+        assert(stop_sent_ > 0);
+    }
+    void         stop_sent_dec(int val)
+    {
+#ifdef GU_DEBUG_MUTEX
+        assert(gu_mutex_owned(&fc_lock));
+#endif
+        assert(stop_sent_ >= val);
+        assert(stop_sent_ > 0);
+        stop_sent_ -= val;
+    }
     long         stop_count;          // counts stop requests received
     long         queue_len;           // slave queue length
     long         upper_limit;         // upper slave queue limit
     long         lower_limit;         // lower slave queue limit
     long         fc_offset;           // offset for catchup phase
     gcs_conn_state_t max_fc_state;    // maximum state when FC is enabled
-    long         stats_fc_sent;       // FC stats counters
+    long         stats_fc_stop_sent;  // FC stats counters
+    long         stats_fc_cont_sent;  //
     long         stats_fc_received;   //
     gcs_fc_t     stfc; // state transfer FC object
 
     /* #603, #606 join control */
-    bool        volatile need_to_join;
     gcs_seqno_t volatile join_seqno;
+    bool        volatile need_to_join;
 
     /* sync control */
     bool         sync_sent_;
@@ -356,8 +381,8 @@ gcs_init (gcs_conn_t* conn, gcs_seqno_t seqno, const uint8_t uuid[GU_UUID_LEN])
  * @param warning warning to log if necessary
  * @return        0 if error can be ignored, original err value if not
  */
-static long
-gcs_check_error (long err, const char* warning)
+static int
+gcs_check_error (int err, const char* warning)
 {
     switch (err)
     {
@@ -388,7 +413,7 @@ gcs_fc_stop_begin (gcs_conn_t* conn)
     long err = 0;
 
     bool ret = (conn->stop_count <= 0                                     &&
-                conn->stop_sent  <= 0                                     &&
+                conn->stop_sent_ <= 0                                     &&
                 conn->queue_len  >  (conn->upper_limit + conn->fc_offset) &&
                 conn->state      <= conn->max_fc_state                    &&
                 !(err = gu_mutex_lock (&conn->fc_lock)));
@@ -398,29 +423,43 @@ gcs_fc_stop_begin (gcs_conn_t* conn)
             abort();
     }
 
-    conn->stop_sent += ret;
-
     return ret;
 }
 
 /* Complement to gcs_fc_stop_begin. */
-static inline long
+static inline int
 gcs_fc_stop_end (gcs_conn_t* conn)
 {
-    long ret;
+#ifdef GU_DEBUG_MUTEX
+    assert(gu_mutex_owned(&conn->fc_lock));
+#endif
 
-    gu_debug ("SENDING FC_STOP (local seqno: %lld, fc_offset: %ld)",
-              conn->local_act_id, conn->fc_offset);
+    int ret = 0;
 
-    ret = gcs_send_fc_event (conn, GCS_FC_STOP);
+    if (conn->stop_sent() <= 0)
+    {
+        conn->stop_sent_inc(1);
+        gu_mutex_unlock (&conn->fc_lock);
 
-    if (ret >= 0) {
-        ret = 0;
-        conn->stats_fc_sent++;
+        ret = gcs_send_fc_event (conn, GCS_FC_STOP);
+
+        gu_mutex_lock (&conn->fc_lock);
+        if (ret >= 0) {
+            ret = 0;
+            conn->stats_fc_stop_sent++;
+        }
+        else {
+            assert (conn->stop_sent() > 0);
+            /* restore counter */
+            conn->stop_sent_dec(1);
+        }
+
+        gu_debug ("SENDING FC_STOP (local seqno: %lld, fc_offset: %ld): %d",
+                 conn->local_act_id, conn->fc_offset, ret);
     }
-    else {
-        conn->stop_sent--;
-        assert (conn->stop_sent >= 0);
+    else
+    {
+        gu_debug ("SKIPPED FC_STOP sending: stop_sent = %d", conn->stop_sent());
     }
 
     gu_mutex_unlock (&conn->fc_lock);
@@ -439,7 +478,7 @@ gcs_fc_cont_begin (gcs_conn_t* conn)
     bool queue_decreased = (conn->fc_offset > conn->queue_len &&
                             (conn->fc_offset = conn->queue_len, true));
 
-    bool ret = (conn->stop_sent    >  0                                   &&
+    bool ret = (conn->stop_sent_  >  0                                    &&
                 (conn->lower_limit >= conn->queue_len || queue_decreased) &&
                 conn->state        <= conn->max_fc_state                  &&
                 !(err = gu_mutex_lock (&conn->fc_lock)));
@@ -449,27 +488,41 @@ gcs_fc_cont_begin (gcs_conn_t* conn)
         abort();
     }
 
-    conn->stop_sent -= ret; // decrement optimistically to allow for parallel
-                            // recv threads
     return ret;
 }
 
 /* Complement to gcs_fc_cont_begin() */
-static inline long
+static inline int
 gcs_fc_cont_end (gcs_conn_t* conn)
 {
-    long ret;
+    int ret = 0;
 
-    assert (GCS_CONN_DONOR >= conn->state);
+    assert (GCS_CONN_JOINER >= conn->state);
 
-    gu_debug ("SENDING FC_CONT (local seqno: %lld, fc_offset: %ld)",
-              conn->local_act_id, conn->fc_offset);
+    if (conn->stop_sent())
+    {
+        conn->stop_sent_dec(1);
+        gu_mutex_unlock (&conn->fc_lock);
 
-    ret = gcs_send_fc_event (conn, GCS_FC_CONT);
+        ret = gcs_send_fc_event (conn, GCS_FC_CONT);
 
-    if (gu_likely (ret >= 0)) { ret = 0; }
+        gu_mutex_lock (&conn->fc_lock);
+        if (gu_likely (ret >= 0)) {
+            ret = 0;
+            conn->stats_fc_cont_sent++;
+        }
+        else {
+            /* restore counter */
+            conn->stop_sent_inc(1);
+        }
 
-    conn->stop_sent += (ret != 0); // fix count in case of error
+        gu_debug ("SENDING FC_CONT (local seqno: %lld, fc_offset: %ld): %d",
+                 conn->local_act_id, conn->fc_offset, ret);
+    }
+    else
+    {
+        gu_debug ("SKIPPED FC_CONT sending: stop_sent = %d", conn->stop_sent());
+    }
 
     gu_mutex_unlock (&conn->fc_lock);
 
@@ -608,19 +661,18 @@ gcs_set_pkt_size (gcs_conn_t *conn, long pkt_size)
     return ret;
 }
 
-static long
+static int
 _release_flow_control (gcs_conn_t* conn)
 {
     int err = 0;
 
     if (gu_unlikely(err = gu_mutex_lock (&conn->fc_lock))) {
-        gu_fatal ("Mutex lock failed: %d (%s)", err, strerror(err));
+        gu_fatal ("FC mutex lock failed: %d (%s)", err, strerror(err));
         abort();
     }
 
-    if (conn->stop_sent) {
-        assert (1 == conn->stop_sent);
-        conn->stop_sent--;
+    if (conn->stop_sent()) {
+        assert (1 == conn->stop_sent());
         err = gcs_fc_cont_end (conn);
     }
     else {
@@ -633,13 +685,20 @@ _release_flow_control (gcs_conn_t* conn)
 static void
 gcs_become_primary (gcs_conn_t* conn)
 {
+    assert(conn->join_seqno == GCS_SEQNO_ILL ||
+           conn->state == GCS_CONN_JOINER    ||
+           conn->state == GCS_CONN_OPEN /* joiner that has received NON_PRIM */);
+
     if (!gcs_shift_state (conn, GCS_CONN_PRIMARY)) {
         gu_fatal ("Protocol violation, can't continue");
         gcs_close (conn);
         abort();
     }
 
-    long ret;
+    conn->join_seqno   = GCS_SEQNO_NIL;
+    conn->need_to_join = false;
+
+    int ret;
 
     if ((ret = _release_flow_control (conn))) {
         gu_fatal ("Failed to release flow control: %ld (%s)",
@@ -704,20 +763,22 @@ gcs_become_donor (gcs_conn_t* conn)
     return 0; // do not pass to application
 }
 
-static long
+static int
 _release_sst_flow_control (gcs_conn_t* conn)
 {
-    long ret = 0;
+    int ret = 0;
 
     do {
-        if (conn->stop_sent > 0) {
-            ret = gcs_send_fc_event (conn, GCS_FC_CONT);
-            conn->stop_sent -= (ret >= 0);
+        ret = gu_mutex_lock(&conn->fc_lock);
+        if (!ret) {
+            ret = gcs_fc_cont_end(conn);
+        }
+        else {
+            gu_fatal("failed to lock FC mutex");
+            abort();
         }
     }
-    while (ret < 0 && -EAGAIN == ret); // we need to send CONT here at all costs
-
-    ret = gcs_check_error (ret, "Failed to release SST flow control.");
+    while (-EAGAIN == ret); // we need to send CONT here at all costs
 
     return ret;
 }
@@ -725,7 +786,7 @@ _release_sst_flow_control (gcs_conn_t* conn)
 static void
 gcs_become_joined (gcs_conn_t* conn)
 {
-    long ret;
+    int ret;
 
     if (GCS_CONN_JOINER == conn->state) {
         ret = _release_sst_flow_control (conn);
@@ -740,6 +801,7 @@ gcs_become_joined (gcs_conn_t* conn)
     /* See also gcs_handle_act_conf () for a case of cluster bootstrapping */
     if (gcs_shift_state (conn, GCS_CONN_JOINED)) {
         conn->fc_offset    = conn->queue_len;
+        conn->join_seqno   = GCS_SEQNO_NIL;
         conn->need_to_join = false;
         gu_debug("Become joined, FC offset %ld", conn->fc_offset);
         /* One of the cases when the node can become SYNCED */
@@ -818,12 +880,12 @@ _reset_pkt_size(gcs_conn_t* conn)
     }
 }
 
-static long
-_join (gcs_conn_t* conn, gcs_seqno_t seqno)
+static int
+s_join (gcs_conn_t* conn)
 {
-    long err;
+    int err;
 
-    while (-EAGAIN == (err = gcs_core_send_join (conn->core, seqno)))
+    while (-EAGAIN == (err = gcs_core_send_join (conn->core, conn->join_seqno)))
         usleep (10000);
 
     switch (err)
@@ -831,6 +893,7 @@ _join (gcs_conn_t* conn, gcs_seqno_t seqno)
     case -ENOTCONN:
         gu_warn ("Sending JOIN failed: %d (%s). "
                  "Will retry in new primary component.", err, strerror(-err));
+        // fall through
     case 0:
         return 0;
     default:
@@ -854,7 +917,10 @@ gcs_handle_act_conf (gcs_conn_t* conn, const void* action)
     {
         /* reset flow control as membership is most likely changed */
         if (!gu_mutex_lock (&conn->fc_lock)) {
-            conn->stop_sent   = 0;
+            /* wake up send monitor if it was paused */
+            if (conn->stop_count > 0) gcs_sm_continue(conn->sm);
+
+            conn->stop_sent_  = 0;
             conn->stop_count  = 0;
             conn->conf_id     = conf->conf_id;
             conn->memb_num    = conf->memb_num;
@@ -869,9 +935,6 @@ gcs_handle_act_conf (gcs_conn_t* conn, const void* action)
         }
 
         conn->sync_sent(false);
-
-        // need to wake up send monitor if it was paused during CC
-        gcs_sm_continue(conn->sm);
     }
     gu_fifo_release (conn->recv_q);
 
@@ -951,7 +1014,7 @@ gcs_handle_act_conf (gcs_conn_t* conn, const void* action)
         /* #603, #606 - duplicate JOIN msg in case we lost it */
         assert (conf->conf_id >= 0);
 
-        if (conn->need_to_join) _join (conn, conn->join_seqno);
+        if (conn->need_to_join) s_join (conn);
 
         break;
     default:
@@ -1094,14 +1157,13 @@ _check_recv_queue_growth (gcs_conn_t* conn, ssize_t size)
 
     if (pause > 0) {
         /* replication needs throttling */
-        if (conn->stop_sent <= 0) {
-            if ((ret = gcs_send_fc_event (conn, GCS_FC_STOP)) >= 0) {
-                conn->stop_sent++;
-                ret = 0;
-            }
-            else {
-                ret = gcs_check_error (ret, "Failed to send SST FC_STOP.");
-            }
+        ret = gu_mutex_lock(&conn->fc_lock);
+        if (!ret) {
+            ret = gcs_fc_stop_end(conn);
+        }
+        else {
+            gu_fatal("failed to lock FC mutex");
+            abort();
         }
 
         if (gu_likely(pause != GU_TIME_ETERNITY)) {
@@ -1216,22 +1278,31 @@ static void *gcs_recv_thread (void *arg)
 
         if (gu_unlikely(ret <= 0)) {
 
-            if (-ETIMEDOUT == ret && _handle_timeout(conn)) continue;
+            gu_debug ("gcs_core_recv returned %d: %s", ret, strerror(-ret));
 
-            struct gcs_recv_act* err_act =
-                (struct gcs_recv_act*) gu_fifo_get_tail(conn->recv_q);
+            if (-ETIMEDOUT == ret && _handle_timeout(conn)) continue;
 
             assert (NULL          == rcvd.act.buf);
             assert (0             == rcvd.act.buf_len);
-            assert (GCS_ACT_ERROR == rcvd.act.type);
+            assert (GCS_ACT_ERROR == rcvd.act.type ||
+                    GCS_ACT_INCONSISTENCY == rcvd.act.type);
             assert (GCS_SEQNO_ILL == rcvd.id);
 
+            if (GCS_ACT_INCONSISTENCY == rcvd.act.type) {
+                /* In the case of inconsistency our concern is to report it to
+                 * replicator ASAP. Current contents of the slave queue are
+                 * meaningless. */
+                gu_fifo_clear(conn->recv_q);
+            }
+
+            struct gcs_recv_act* err_act =
+                (struct gcs_recv_act*) gu_fifo_get_tail(conn->recv_q);
+
             err_act->rcvd     = rcvd;
             err_act->local_id = GCS_SEQNO_ILL;
 
             GCS_FIFO_PUSH_TAIL (conn, rcvd.act.buf_len);
 
-            gu_debug ("gcs_core_recv returned %d: %s", ret, strerror(-ret));
             break;
         }
 
@@ -1297,12 +1368,12 @@ static void *gcs_recv_thread (void *arg)
                 recv_act->local_id = this_act_id;
 
                 conn->queue_len = gu_fifo_length (conn->recv_q) + 1;
-                bool send_stop  = gcs_fc_stop_begin (conn);
+                bool const send_stop(gcs_fc_stop_begin(conn));
 
                 // release queue
                 GCS_FIFO_PUSH_TAIL (conn, rcvd.act.buf_len);
 
-                if (gu_unlikely(GCS_CONN_JOINER == conn->state)) {
+                if (gu_unlikely(GCS_CONN_JOINER == conn->state && !send_stop)) {
                     ret = _check_recv_queue_growth (conn, rcvd.act.buf_len);
                     assert (ret <= 0);
                     if (ret < 0) break;
@@ -1544,9 +1615,9 @@ long gcs_interrupt (gcs_conn_t* conn, long handle)
     return gcs_sm_interrupt (conn->sm, handle);
 }
 
-gcs_seqno_t gcs_caused(gcs_conn_t* conn)
+long gcs_caused(gcs_conn_t* conn, gcs_seqno_t& seqno)
 {
-    return gcs_core_caused(conn->core);
+    return gcs_core_caused(conn->core, seqno);
 }
 
 /* Puts action in the send queue and returns after it is replicated */
@@ -1943,12 +2014,18 @@ gcs_set_last_applied (gcs_conn_t* conn, gcs_seqno_t seqno)
 }
 
 long
-gcs_join (gcs_conn_t* conn, gcs_seqno_t seqno)
+gcs_join (gcs_conn_t* conn, gcs_seqno_t const seqno)
 {
-    conn->join_seqno   = seqno;
-    conn->need_to_join = true;
+    assert(conn->join_seqno <= seqno || seqno < 0);
 
-    return _join (conn, seqno);
+    if (seqno < 0 || seqno >= conn->join_seqno)
+    {
+        conn->join_seqno   = seqno;
+        conn->need_to_join = true;
+        return s_join (conn);
+    }
+
+    return 0;
 }
 
 gcs_seqno_t gcs_local_sequence(gcs_conn_t* conn)
@@ -1975,7 +2052,8 @@ gcs_get_stats (gcs_conn_t* conn, struct gcs_stats* stats)
                       &stats->fc_paused_ns,
                       &stats->fc_paused_avg);
 
-    stats->fc_sent     = conn->stats_fc_sent;
+    stats->fc_ssent    = conn->stats_fc_stop_sent;
+    stats->fc_csent    = conn->stats_fc_cont_sent;
     stats->fc_received = conn->stats_fc_received;
 }
 
@@ -1984,8 +2062,9 @@ gcs_flush_stats(gcs_conn_t* conn)
 {
     gu_fifo_stats_flush(conn->recv_q);
     gcs_sm_stats_flush (conn->sm);
-    conn->stats_fc_sent     = 0;
-    conn->stats_fc_received = 0;
+    conn->stats_fc_stop_sent = 0;
+    conn->stats_fc_cont_sent = 0;
+    conn->stats_fc_received  = 0;
 }
 
 void gcs_get_status(gcs_conn_t* conn, gu::Status& status)
@@ -2223,6 +2302,12 @@ long gcs_param_set  (gcs_conn_t* conn, const char* key, const char *value)
     else if (!strcmp (key, GCS_PARAMS_MAX_THROTTLE)) {
         return _set_max_throttle (conn, value);
     }
+#ifdef GCS_SM_DEBUG
+    else if (!strcmp (key, GCS_PARAMS_SM_DUMP)) {
+        gcs_sm_dump_state(conn->sm, stderr);
+        return 0;
+    }
+#endif /* GCS_SM_DEBUG */
     else {
         return gcs_core_param_set (conn->core, key, value);
     }
