@@ -89,6 +89,9 @@ fi
 if ! which "$CC" ; then echo "Can't execute $CC" ; exit 1; fi
 if ! which "$CXX"; then echo "Can't execute $CXX"; exit 1; fi
 
+CFLAGS=${CFLAGS:-""}
+CXXFLAGS=${CXXFLAGS:-""}
+
 export CC CXX LD_LIBRARY_PATH
 
 EXTRA_SYSROOT=${EXTRA_SYSROOT:-""}
@@ -456,8 +459,10 @@ then
                 BOOST_OPT="-DWITH_BOOST=boost_$MYSQL_MM_VER"
                 [ "yes" = "$BOOTSTRAP" ] && \
                     BOOST_OPT="$BOOST_OPT -DDOWNLOAD_BOOST=1"
+                SSL_OPT="-DWITH_SSL=yes"
             else
                 BOOST_OPT=""
+                SSL_OPT="-DWITH_SSL=yes"
             fi
 
             if [ "$MYSQL_BUILD_DIR" != "$MYSQL_SRC" ]
@@ -477,12 +482,14 @@ then
             cmake \
                   -DCMAKE_C_COMPILER=$(basename $CC) \
                   -DCMAKE_CXX_COMPILER=$(basename $CXX) \
+                  -DCMAKE_C_FLAGS="$CFLAGS" \
+                  -DCMAKE_CXX_FLAGS="$CXXFLAGS" \
                   -DBUILD_CONFIG=mysql_release \
                   "${CMAKE_LAYOUT_OPTIONS[@]}" \
                   $BUILD_OPT \
                   -DWITH_WSREP=1 \
                   -DWITH_EXTRA_CHARSETS=all \
-                  -DWITH_SSL=yes \
+                  $SSL_OPT \
                   -DWITH_ZLIB=system \
                   -DMYSQL_MAINTAINER_MODE=0 \
                   $MEMCACHED_OPT \
